"""Main entry point for the formatter."""
from __future__ import annotations

import difflib
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, Sequence

from pyproject_fmt.cli import PyProjectFmtNamespace, cli_args
from pyproject_fmt.formatter import format_pyproject

if TYPE_CHECKING:
    from pyproject_fmt import Config

GREEN = "\u001b[32m"
RED = "\u001b[31m"
RESET = "\u001b[0m"


def color_diff(diff: Iterable[str]) -> Iterable[str]:
    """
    Visualize difference with colors.

    :param diff: the diff lines
    """
    for line in diff:
        if line.startswith("+"):
            yield f"{GREEN}{line}{RESET}"
        elif line.startswith("-"):
            yield f"{RED}{line}{RESET}"
        else:
            yield line


def _handle_one(config: Config, opts: PyProjectFmtNamespace) -> bool:
    formatted = format_pyproject(config)
    before = config.toml
    changed = before != formatted
    if opts.stdout:  # stdout just prints new format to stdout
        print(formatted, end="")  # noqa: T201
        return changed

    if before != formatted and not opts.check:
        config.pyproject_toml.write_text(formatted, encoding="utf-8")
    try:
        name = str(config.pyproject_toml.relative_to(Path.cwd()))
    except ValueError:
        name = str(config.pyproject_toml)
    diff: Iterable[str] = []
    if changed:
        diff = difflib.unified_diff(before.splitlines(), formatted.splitlines(), fromfile=name, tofile=name)

    if diff:
        diff = color_diff(diff)
        print("\n".join(diff))  # print diff on change  # noqa: T201
    else:
        print(f"no change for {name}")  # noqa: T201
    return changed


def run(args: Sequence[str] | None = None) -> int:
    """
    Run the formatter.

    :param args: CLI arguments
    :return: exit code
    """
    opts = cli_args(sys.argv[1:] if args is None else args)
    results = [_handle_one(config, opts) for config in opts.configs]
    return 1 if any(results) else 0  # exit with non success on change


if __name__ == "__main__":
    raise SystemExit(run())



"""Format your pyproject.toml."""
from __future__ import annotations

from ._version import version as __version__
from .formatter import format_pyproject
from .formatter.config import Config

__all__ = [
    "__version__",
    "Config",
    "format_pyproject",
]



"""Format the project table."""
from __future__ import annotations

import re
from shutil import which
from subprocess import CalledProcessError, check_output
from typing import TYPE_CHECKING, Optional, cast

from packaging.specifiers import SpecifierSet
from packaging.utils import canonicalize_name
from packaging.version import Version
from tomlkit.items import Array, String, Table, Trivia

from .pep508 import normalize_pep508_array
from .util import ensure_newline_at_end, order_keys, sorted_array

if TYPE_CHECKING:
    from tomlkit.toml_document import TOMLDocument

    from .config import Config

_PY_MIN_VERSION: int = 7


def fmt_project(parsed: TOMLDocument, conf: Config) -> None:  # noqa: C901
    """
    Format the project table.

    :param parsed: the raw parsed table
    :param conf: configuration
    """
    project = cast(Optional[Table], parsed.get("project"))
    if project is None:
        return

    if "name" in project:  # normalize names to hyphen so sdist / wheel have the same prefix
        name = project["name"]
        assert isinstance(name, str)  # noqa: S101
        project["name"] = canonicalize_name(name)
    if "description" in project:
        project["description"] = String.from_raw(str(project["description"]).strip())

    sorted_array(cast(Optional[Array], project.get("keywords")), indent=conf.indent)
    sorted_array(cast(Optional[Array], project.get("dynamic")), indent=conf.indent)

    if "requires-python" in project:
        _add_py_classifiers(project, py_max_version=conf.max_supported_python)

    sorted_array(cast(Optional[Array], project.get("classifiers")), indent=conf.indent, custom_sort="natsort")

    normalize_pep508_array(
        requires_array=cast(Optional[Array], project.get("dependencies")),
        indent=conf.indent,
        keep_full_version=conf.keep_full_version,
    )
    if "optional-dependencies" in project:
        opt_deps = cast(Table, project["optional-dependencies"])
        for value in opt_deps.values():
            normalize_pep508_array(
                requires_array=cast(Array, value),
                indent=conf.indent,
                keep_full_version=conf.keep_full_version,
            )
        order_keys(opt_deps, (), sort_key=lambda k: k[0])  # pragma: no branch

    for of_type in ("scripts", "gui-scripts", "entry-points", "urls"):
        if of_type in project:
            table = cast(Table, project[of_type])
            order_keys(table, (), sort_key=lambda k: k[0])  # pragma: no branch

    if "entry-points" in project:  # order entry points sub-table
        entry_points = cast(Table, project["entry-points"])
        order_keys(entry_points, (), sort_key=lambda k: k[0])  # pragma: no branch
        for entry_point in entry_points.values():
            order_keys(entry_point, (), sort_key=lambda k: k[0])  # pragma: no branch

    # order maintainers and authors table
    # handle readme table

    key_order = [
        "name",
        "version",
        "description",
        "readme",
        "keywords",
        "license",
        "license-files",
    ]
    key_order.extend(
        [
            "maintainers",
            "authors",
            "requires-python",
            "classifiers",
            "dynamic",
            "dependencies",
        ],
    )
    # these go at the end as they may be inline or exploded
    key_order.extend(
        ["optional-dependencies", "urls", "scripts", "gui-scripts", "entry-points"],
    )
    order_keys(project, key_order)
    ensure_newline_at_end(project)


def _add_py_classifiers(project: Table, *, py_max_version: Version) -> None:
    specifiers = SpecifierSet(project.get("requires-python", f">=3.{_PY_MIN_VERSION}"))

    min_version = _get_min_version_classifier(specifiers)
    max_version = _get_max_version_classifier(specifiers, py_max_version=py_max_version)

    allowed_versions = list(specifiers.filter(f"3.{v}" for v in range(min_version, max_version + 1)))

    add = [f"Programming Language :: Python :: {v}" for v in allowed_versions]
    add.append("Programming Language :: Python :: 3 :: Only")

    if "classifiers" in project:
        classifiers: Array = cast(Array, project["classifiers"])
    else:
        classifiers = Array([], Trivia(), multiline=False)
        project["classifiers"] = classifiers

    exist = set(classifiers.unwrap())
    remove = [e for e in exist if re.fullmatch(r"Programming Language :: Python :: \d.*", e) and e not in add]
    deleted = 0
    for at, item in enumerate(list(classifiers)):
        if item in remove:
            del classifiers[at - deleted]
            deleted += 1

    for entry in add:
        if entry not in classifiers:
            classifiers.insert(len(add), entry)


def _get_min_version_classifier(specifiers: SpecifierSet) -> int:
    min_version: list[int] = []
    for specifier in specifiers:
        if specifier.operator == ">=":
            min_version.append(Version(specifier.version).minor)
        if specifier.operator == ">":
            min_version.append(Version(specifier.version).minor + 1)
    return min(min_version, default=_PY_MIN_VERSION)


def _get_max_version_classifier(specifiers: SpecifierSet, *, py_max_version: Version) -> int:
    max_version: list[int] = []

    for specifier in specifiers:
        if specifier.operator == "<=":
            max_version.append(Version(specifier.version).minor)
        if specifier.operator == "<":
            max_version.append(Version(specifier.version).minor - 1)

    return max(max_version) if max_version else (_get_max_version_tox() or py_max_version.minor)


def _get_max_version_tox() -> int | None:
    tox = which("tox")
    if tox is not None:  # pragma: no branch
        try:
            tox_environments = check_output([tox, "-aqq"], encoding="utf-8", text=True)  # noqa: S603
        except (OSError, CalledProcessError):
            tox_environments = ""
        if not re.match(r"ROOT: No .* found, assuming empty", tox_environments):
            found = set()
            for env in tox_environments.split():
                for part in env.split("-"):
                    match = re.match(r"py(\d)(\d+)", part)
                    if match:
                        found.add(int(match.groups()[1]))
            if found:
                return max(found)
    return None


__all__ = [
    "fmt_project",
]



"""Defines configuration for the formatter."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from packaging.version import Version

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Final


DEFAULT_INDENT: Final[int] = 2  #: default indentation level
DEFAULT_MAX_SUPPORTED_PYTHON: Final[str] = "3.12"  #: default maximum supported Python version


@dataclass(frozen=True)
class Config:
    """Configuration flags for the formatting."""

    pyproject_toml: Path
    toml: str  #: the text to format
    indent: int = DEFAULT_INDENT  #: indentation to apply
    keep_full_version: bool = False
    max_supported_python: Version = field(default_factory=lambda: Version(DEFAULT_MAX_SUPPORTED_PYTHON))


__all__ = [
    "Config",
    "DEFAULT_INDENT",
    "DEFAULT_MAX_SUPPORTED_PYTHON",
]
